from collections import Counter

# import numpy as np
from PySide6.QtCore import (
    QPoint,
    QRect,
    Qt,
    Signal,
)
from PySide6.QtGui import (
    QColor,
    QKeyEvent,
    QPainter,
    QWheelEvent,
)
from PySide6.QtWidgets import (
    QGraphicsView,
)

# import pdk.symLayers as symlyr
import pdk.schLayers as schlyr
import revedaEditor.common.net as net
import revedaEditor.backend.undoStack as us


# import os
# if os.environ.get('REVEDASIM_PATH'):
#     import revedasim.simMainWindow as smw

class editorView(QGraphicsView):
    """
    The qgraphicsview for qgraphicsscene. It is used for both schematic and layout editors.
    """
    zoomFactorChanged = Signal(float)
    def __init__(self, scene, parent):
        super().__init__(scene, parent)
        self.parent = parent
        self.editor = self.parent.parent
        self.scene = scene
        self.logger = self.scene.logger
        self.majorGrid = self.editor.majorGrid
        self.snapGrid = self.editor.snapGrid
        self.snapTuple = self.editor.snapTuple
        self.gridbackg = True
        self.linebackg = False
        self._left: QPoint = QPoint(0, 0)
        self._right: QPoint = QPoint(0, 0)
        self._top: QPoint = QPoint(0, 0)
        self._bottom: QPoint = QPoint(0, 0)
        self.viewRect = QRect(0, 0, 0, 0)
        self.zoomFactor = 1.0
        self.init_UI()

    def init_UI(self):
        """
        Initializes the user interface.

        This function sets up various properties of the QGraphicsView object, such as rendering hints,
        mouse tracking, transformation anchors, and cursor shape.

        Returns:
            None
        """
        self.setViewportUpdateMode(QGraphicsView.FullViewportUpdate)
        # self.setCacheMode(QGraphicsView.CacheBackground)
        self.setMouseTracking(True)
        self.setTransformationAnchor(QGraphicsView.AnchorUnderMouse)
        self.setResizeAnchor(QGraphicsView.AnchorUnderMouse)
        self.setInteractive(True)
        self.setCursor(Qt.CrossCursor)
        self.setRenderHint(QPainter.Antialiasing, True)
        self.setRenderHint(QPainter.TextAntialiasing, True)
        self.viewRect = self.mapToScene(self.rect()).boundingRect().toRect()


    def wheelEvent(self, event: QWheelEvent) -> None:
        """
        Handle the wheel event for zooming in and out of the view.

        Args:
            event (QWheelEvent): The wheel event to handle.
        """
        # Get the current center point of the view
        oldPos = self.mapToScene(self.viewport().rect().center())

        # Perform the zoom
        self.zoomFactor = 1.2 if event.angleDelta().y() > 0 else 1 / 1.2
        self.scale(self.zoomFactor, self.zoomFactor)

        # Get the new center point of the view
        newPos = self.mapToScene(self.viewport().rect().center())

        # Calculate the delta and adjust the scene position
        delta = newPos - oldPos
        self.translate(delta.x(), delta.y())
        self.zoomFactorChanged.emit(self.zoomFactor)

    def snapToBase(self, number, base):
        """
        Restrict a number to the multiples of base
        """
        return int(base * int(round(number / base)))

    def snapToGrid(self, point: QPoint, snapTuple: tuple[int, int]):
        """
        snap point to grid. Divides and multiplies by grid size.
        """
        return QPoint(
            self.snapToBase(point.x(), snapTuple[0]),
            self.snapToBase(point.y(), snapTuple[1]),
        )

    def drawBackground(self, painter, rect):
        """
        Draws the background of the painter within the given rectangle.

        Args:
            painter (QPainter): The painter object to draw on.
            rect (QRect): The rectangle to draw the background within.
        """

        # Fill the rectangle with black color
        painter.fillRect(rect, QColor("black"))

        # Calculate the coordinates of the left, top, bottom, and right edges of the rectangle
        self._left = int(rect.left()) - (int(rect.left()) % self.majorGrid)
        self._top = int(rect.top()) - (int(rect.top()) % self.majorGrid)
        self._bottom = int(rect.bottom())
        self._right = int(rect.right())

        if self.gridbackg:
            # Set the pen color to gray
            painter.setPen(QColor("white"))

            # Create a range of x and y coordinates for drawing the grids
            x_coords, y_coords = self.findCoords()

            for x_coord in x_coords:
                for y_coord in y_coords:
                    painter.drawPoint(x_coord, y_coord)

        elif self.linebackg:
            # Set the pen color to gray
            painter.setPen(QColor("gray"))

            # Create a range of x and y coordinates for drawing the lines
            x_coords, y_coords = self.findCoords()

            # Draw vertical lines
            for x in x_coords:
                painter.drawLine(x, self._top, x, self._bottom)

            # Draw horizontal lines
            for y in y_coords:
                painter.drawLine(self._left, y, self._right, y)

        else:
            # Call the base class method to draw the background
            super().drawBackground(painter, rect)

    def findCoords(self):
        """
        Calculate the coordinates for drawing lines or points on a grid.

        Returns:
            tuple: A tuple containing the x and y coordinates for drawing the lines or points.
        """
        x_coords = range(self._left, self._right, self.majorGrid)
        y_coords = range(self._top, self._bottom, self.majorGrid)

        if 160 <= len(x_coords) < 320:
            # Create a range of x and y coordinates for drawing the lines
            x_coords = range(self._left, self._right, self.majorGrid * 2)
            y_coords = range(self._top, self._bottom, self.majorGrid * 2)
        elif 320 <= len(x_coords) < 640:
            x_coords = range(self._left, self._right, self.majorGrid * 4)
            y_coords = range(self._top, self._bottom, self.majorGrid * 4)
        elif 640 <= len(x_coords) < 1280:
            x_coords = range(self._left, self._right, self.majorGrid * 8)
            y_coords = range(self._top, self._bottom, self.majorGrid * 8)
        elif 1280 <= len(x_coords) < 2560:
            x_coords = range(self._left, self._right, self.majorGrid * 16)
            y_coords = range(self._top, self._bottom, self.majorGrid * 16)
        elif len(x_coords) >= 2560:  # grid dots are too small to see
            x_coords = range(self._left, self._right, self.majorGrid * 1000)
            y_coords = range(self._top, self._bottom, self.majorGrid * 1000)

        return x_coords, y_coords

    def keyPressEvent(self, event: QKeyEvent):
        match event.key():
            case Qt.Key_F:
                self.scene.fitItemsInView()
            case Qt.Key_Left:
                self.scene.moveSceneLeft()
            case Qt.Key_Right:
                self.scene.moveSceneRight()
            case Qt.Key_Up:
                self.scene.moveSceneUp()
            case Qt.Key_Down:
                self.scene.moveSceneDown()
            case Qt.Key_Escape:
                self.scene.editModes.setMode("selectItem")
                self.editor.messageLine.setText("Select Item")
            case _:
                super().keyPressEvent(event)

    def printView(self, printer):
        """
        Print view using selected Printer.

        Args:
            printer (QPrinter): The printer object to use for printing.

        This method prints the current view using the provided printer. It first creates a QPainter object
        using the printer. Then, it toggles the gridbackg and linebackg attributes. After that, it calls
        the revedaPrint method to render the view onto the painter. Finally, it toggles the gridbackg
        and linebackg attributes back to their original state.
        """
        painter = QPainter(printer)

        # Toggle gridbackg attribute
        if self.gridbackg:
            self.gridbackg = False
        else:
            self.linebackg = False

        # Render the view onto the painter
        self.revedaPrint(painter)

        # Toggle gridbackg and linebackg attributes back to their original state
        self.gridbackg = not self.gridbackg
        self.linebackg = not self.linebackg

    def revedaPrint(self, painter):
        viewport_geom = self.viewport().geometry()
        self.drawBackground(painter, viewport_geom)
        painter.drawText(viewport_geom, "Revolution EDA")
        self.render(painter)
        painter.end()


class symbolView(editorView):
    def __init__(self, scene, parent):
        self.scene = scene
        self.parent = parent
        super().__init__(self.scene, self.parent)

    def keyPressEvent(self, event: QKeyEvent):
        match event.key():
            case Qt.Key_Escape:
                self.scene.editModes.setMode("selectItem")
                self.editor.messageLine.setText("Select Item")
                if self.scene.polygonGuideLine is not None:
                    self.scene.removeItem(self.scene.polygonGuideLine)
                    self.scene.polygonGuideLine = None
                    self.scene.newPolygon = None
            case _:
                super().keyPressEvent(event)


class schematicView(editorView):
    def __init__(self, scene, parent):
        self.scene = scene
        self.parent = parent
        super().__init__(self.scene, self.parent)
        self._dotRadius = 2

    def mouseReleaseEvent(self, event):
        super().mouseReleaseEvent(event)
        self.viewRect = self.mapToScene(self.rect()).boundingRect().toRect()
        viewSnapLinesSet = {
            guideLineItem
            for guideLineItem in self.scene.items(self.viewRect)
            if isinstance(guideLineItem, net.guideLine)
        }
        self.removeSnapLines(viewSnapLinesSet)

        self.mergeSplitViewNets()

    def mergeSplitViewNets(self):
        netsInView = [
            netItem
            for netItem in self.scene.items(self.viewRect)
            if isinstance(netItem, net.schematicNet)
        ]
        for netItem in netsInView:
            if netItem.scene():
                self.scene.mergeSplitNets(netItem)

    def removeSnapLines(self, viewSnapLinesSet):

        undoCommandList = []
        for snapLine in viewSnapLinesSet:
            lines: list[net.schematicNet] = self.scene.addStretchWires(
                snapLine.sceneEndPoints[0], snapLine.sceneEndPoints[1]
            )

            if lines:
                for line in lines:
                    line.inheritGuideLine(snapLine)
                undoCommandList.append(us.addShapesUndo(self.scene, lines))

        undoCommandList.append(us.deleteShapesUndo(self.scene, list(viewSnapLinesSet)))
        self.scene.addUndoMacroStack(undoCommandList, 'Stretch Wires')

    def drawBackground(self, painter, rect):
        super().drawBackground(painter, rect)
        xextend = self._right - self._left
        yextend = self._bottom - self._top
        netsInView = [
            netItem
            for netItem in self.scene.items(rect)
            if isinstance(netItem, net.schematicNet)
        ]
        if xextend <= 1000 or yextend <= 1000:
            netEndPoints = []
            for netItem in netsInView:
                netEndPoints.extend(netItem.sceneEndPoints)
            pointCountsDict = Counter(netEndPoints)
            dotPoints = [
                point for point, count in pointCountsDict.items() if count >= 3
            ]
            painter.setPen(schlyr.wirePen)
            painter.setBrush(schlyr.wireBrush)
            for dotPoint in dotPoints:
                painter.drawEllipse(dotPoint, self._dotRadius, self._dotRadius)

    def keyPressEvent(self, event: QKeyEvent):
        """
        Handles the key press event for the editor view.

        Args:
            event (QKeyEvent): The key press event to handle.

        """
        if event.key() == Qt.Key_Escape:
            # Esc key pressed, remove snap rect and reset states
            self.scene.removeSnapRect()
            if self.scene._newNet is not None:
                # New net creation mode, cancel creation
                self.scene._newNet = None
            elif self.scene._stretchNet is not None:
                # Stretch net mode, cancel stretch
                self.scene._stretchNet.setSelected(False)
                self.scene._stretchNet.stretch = False
                self.scene.checkNewNet(self.scene._stretchNet)
            # Set the edit mode to select item
            self.scene.editModes.setMode("selectItem")
        super().keyPressEvent(event)


class layoutView(editorView):
    def __init__(self, scene, parent):
        self.scene = scene
        self.parent = parent
        super().__init__(self.scene, self.parent)


    def keyPressEvent(self, event: QKeyEvent):
        if event.key() == Qt.Key_Escape:
            if self.scene._newPath is not None:
                self.scene._newPath = None
            elif self.scene._newRect:
                self.scene._newRect = None
            elif self.scene._stretchPath is not None:
                self.scene._stretchPath.setSelected(False)
                self.scene._stretchPath.stretch = False
                self.scene._stretchPath = None
            elif self.scene._newPolygon:
                self.scene.removeItem(self.scene._polygonGuideLine)
                self.scene._newPolygon.points.pop(0) # remove first duplicate point
                self.scene._newPolygon = None

            self.scene.editModes.setMode("selectItem")
        super().keyPressEvent(event)



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting or consulting/
#    support services related to the Software), a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#
import pathlib
from pathlib import Path

from PySide6.QtWidgets import (
    QMainWindow,
)
from PySide6.QtCore import (
    QPoint,
    QRect,
)
import revedaEditor.backend.libraryModelView as lmview
import revedaEditor.backend.libraryMethods as libm
import revedaEditor.backend.schBackEnd as scb
import revedaEditor.gui.editorWindows as edw
import revedaEditor.common.shapes as shp
import revedaEditor.common.labels as lbl
import revedaEditor.fileio.symbolEncoder as symenc
import pdk.callbacks as clb

import re
import os


class importXschemSym:
    """
    Imports a xschem sym file
    """

    def __init__(
        self,
        parent: QMainWindow,
        filePathObj: Path,
        libraryView: lmview.designLibrariesView,
        libraryName: str,
    ):
        self.parent = parent
        self.filePathObj = filePathObj
        self.libraryView = libraryView
        self.libraryName = libraryName
        self._scaleFactor = 4.0
        self._labelHeight = 16
        self._labelXOffset = 40
        self._labelYOffset = 16
        self._labelList = []
        self._functions = []
        self._pins = []
        self._tclExpressLines = []
        self._expressionDict = {}
        self.cellName = self.filePathObj.stem
        libItem = libm.getLibItem(self.libraryView.libraryModel, self.libraryName)

        cellItem = scb.createCell(
            self.parent, self.libraryView.libraryModel, libItem, self.cellName
        )
        symbolViewItem = scb.createCellView(self.parent, "symbol", cellItem)
        self.symbolWindow = edw.symbolEditor(
            symbolViewItem, self.parent.libraryDict, self.libraryView
        )
        self.symbolScene = self.symbolWindow.centralW.scene
        
    def importSymFile(self):

        with self.filePathObj.open("r") as file:
            for line in file.readlines():
                lineTokens = line.split()
                if line[0] == "L" and len(lineTokens) > 4:
                    self.symbolScene.lineDraw(
                        QPoint(
                            self._scaleFactor * float(lineTokens[2]),
                            self._scaleFactor * float(lineTokens[3]),
                        ),
                        QPoint(
                            self._scaleFactor * float(lineTokens[4]),
                            self._scaleFactor * float(lineTokens[5]),
                        ),
                    )
                elif line[0] == "B" and len(lineTokens) > 4:
                    properties = self.findProperties(line)
                    if "name" in properties.keys():
                        pin = shp.symbolPin(
                            QPoint(0, 0),
                            properties.get("name", ""),
                            properties.get("dir", "input").capitalize(),
                            shp.symbolPin.pinTypes[0],
                        )
                        pin.rect = QRect(
                            QPoint(
                                self._scaleFactor * float(lineTokens[2]),
                                self._scaleFactor * float(lineTokens[3]),
                            ),
                            QPoint(
                                self._scaleFactor * float(lineTokens[4]),
                                self._scaleFactor * float(lineTokens[5]),
                            ),
                        )
                        self.symbolScene.addItem(pin)
                        self._pins.append(pin)
                    else:
                        self.symbolScene.rectDraw(
                            QPoint(
                                self._scaleFactor * float(lineTokens[2]),
                                self._scaleFactor * float(lineTokens[3]),
                            ),
                            QPoint(
                                self._scaleFactor * float(lineTokens[4]),
                                self._scaleFactor * float(lineTokens[5]),
                            ),
                        )
                elif line[0] == "P" and len(lineTokens) > 4:
                    numberPoints = int(float(lineTokens[2]))
                    points = []
                    for i in range(numberPoints):
                        points.append(
                            QPoint(
                                self._scaleFactor * float(lineTokens[2 * i + 3]),
                                self._scaleFactor * float(lineTokens[2 * i + 4]),
                            )
                        )
                    polygon = shp.symbolPolygon(points)
                    self.symbolScene.addItem(polygon)
                elif line[0] == "T" and len(lineTokens) > 4:
                    if "tcleval" in line:
                    # self._functions.append(line)  # just a reminder
                        self._tclExpressLines.append(line)

        with self.filePathObj.open("r") as file:
            fileContent = file.read()
            pattern = re.compile(r"K\s*{[^}]+}")
            formatStringMatch = pattern.search(fileContent)
            if formatStringMatch:
                formatString = formatStringMatch.group()
                formatDict = self.processFormatString(formatString)
                self.symbolScene.attributeList = list()
                templateDict = formatDict.get("template")
                textLocation = (self._labelXOffset, self._labelYOffset)
                if templateDict:
                    if templateDict.get("model"):
                        self.symbolScene.attributeList.append(
                            symenc.symbolAttribute("modelName", templateDict["model"])
                        )
                        templateDict.pop("model")
                    if templateDict.get("spiceprefix"):
                        self.symbolScene.attributeList.append(
                            symenc.symbolAttribute(
                                "spiceprefix", templateDict["spiceprefix"]
                            )
                        )
                        templateDict.pop("spiceprefix")
                    if templateDict.get("name"):
                        templateDict.pop("name")
                    for key, value in templateDict.items():
                        label = lbl.symbolLabel(
                            QPoint(textLocation[0], textLocation[1]),
                            f"[@{key}:{key.lower()}=%:{key.lower()}={value}]",
                            lbl.symbolLabel.labelTypes[1],
                            self._labelHeight,
                            lbl.symbolLabel.labelAlignments[0],
                            lbl.symbolLabel.labelOrients[0],
                            lbl.symbolLabel.labelUses[1],
                        )
                        textLocation = (
                            textLocation[0],
                            textLocation[1] - self._labelHeight - self._labelYOffset,
                        )
                        label.labelDefs()
                        label.labelVisible = True
                        label.setOpacity(1)
                        self.symbolScene.addItem(label)
                        self._labelList.append(label.labelName)
                    label = lbl.symbolLabel(
                        QPoint(textLocation[0], textLocation[1]),
                        "[@instName]",
                        lbl.symbolLabel.labelTypes[1],
                        self._labelHeight,
                        lbl.symbolLabel.labelAlignments[0],
                        lbl.symbolLabel.labelOrients[0],
                        lbl.symbolLabel.labelUses[1],
                    )
                    label.labelDefs()
                    label.labelVisible = True
                    label.setOpacity(1)
                    self.symbolScene.addItem(label)

                if formatDict.get("format"):
                    netlistLine = (
                        formatDict["format"]
                        .replace("@name", "@instName")
                        .replace("@model", "%modelName")
                        .replace("@spiceprefix", "%spiceprefix")
                        .replace("@pinlist", "@pinList")
                    )
                    self.symbolScene.attributeList.append(
                        symenc.symbolAttribute("XyceSymbolNetlistLine", netlistLine)
                    )
            pinNames = [pin.pinName for pin in self._pins]
            self.symbolScene.attributeList.append(
                symenc.symbolAttribute("pinOrder", ", ".join(pinNames))
            )
        self.processTclEval(self._tclExpressLines)
        self.symbolWindow.checkSaveCell()
        


    @property
    def scaleFactor(self) -> float:
        return self._scaleFactor

    @scaleFactor.setter
    def scaleFactor(self, value: float):
        self._scaleFactor = value

    @staticmethod
    def findProperties(line: str):
        properties = {}
        # Remove curly braces from the string
        propertiesStr = line.split("{")[1].split("}")[0]
        # Split the string by commas to separate key-value pairs
        pairs = propertiesStr.split()
        for pair in pairs:
            key, value = pair.split("=")
            key = key.strip()
            value = value.strip()
            properties[key] = value
        return properties

    def parseTextLine(self, line: str):
        text = line.split("{")[1].split("}")[0]
        textPropertiesStr = line.split("{")[2].split("}")[0]
        pairs = textPropertiesStr.split()
        textProperties = {}
        for pair in pairs:
            key, value = pair.split("=")
            key = key.strip()
            value = value.strip()
            textProperties[key] = value
        restList = line.split("{")[1].split("}")[1].split()

        textLocation = [
            self._scaleFactor * float(restList[0]),
            self._scaleFactor * float(restList[1]),
        ]
        rotationAngle = float(restList[2])
        return text, textProperties, textLocation, rotationAngle

    @staticmethod
    def processFormatString(formatString: str):
        lines = " ".join([line.strip() for line in formatString.strip().split("\n")])
        # print(lines)
        joinedLines = lines.split("{")[1].split("}")[0]
        startIndex = 0
        indexes = []
        while True:
            startIndex = joinedLines.find('="', startIndex)
            if startIndex == -1:
                break
            startIndex += 2
            indexes.append(startIndex)
        startIndex = 0
        # print(f'Keys: {keys}')
        values = []
        for index in indexes:
            endIndex = joinedLines.find('"', index)
            values.append(joinedLines[index:endIndex])

        for valueItem in values:
            joinedLines = joinedLines.replace(valueItem, "")

        keys = [item.replace('=""', "") for item in joinedLines.split()[1:]]
        formatDict = {key: value for key, value in zip(keys, values)}
        templateDict = {
            item.split("=")[0]: item.split("=")[1]
            for item in formatDict["template"].split()
        }
        formatDict["template"] = templateDict
        return formatDict

    def processTclEval(self,inputLines:list[str]):
        tclEvalMatches = []
        parameterMatches = []
        tclPattern = r'\\{(.*?)\\}'
        parameterPattern =  r'tcleval\((.*?)=\['
        for lineItem in inputLines:
            tclEvalMatches.append(re.findall(tclPattern, lineItem)[0])
            parameterMatches.append(re.findall(parameterPattern,lineItem)[0])

        for (key, value) in zip(parameterMatches, tclEvalMatches):
            self._expressionDict[key] = value

        clbPathObj = pathlib.Path(clb.__file__)
        with clbPathObj.open("a") as clbFile:
            clbFile.write("\n\n")
            clbFile.write(f"class {self.cellName}(baseInst):\n")
            clbFile.write(f"    def __init__(self, labels_dict:dict):\n")
            clbFile.write(f"        super().__init__(labels_dict)\n")
            for labelName in self._labelList:
                clbFile.write(f"        self.{labelName[1:]} = Quantity(self._labelsDict["
                              f"'{labelName}'].labelValue)\n")

            for key, value in self._expressionDict.items():
                for labelName in self._labelList:
                    value = value.replace(labelName, f'self.{labelName[1:]}')
                clbFile.write("\n")
                clbFile.write(f"    def {key}parm(self):\n")
                clbFile.write(f"       returnValue = {value}\n")
                clbFile.write(f"       return returnValue\n")
                label = lbl.symbolLabel(
                    QPoint(0, 0),
                    f"{key} = {key}parm()",
                    lbl.symbolLabel.labelTypes[2],
                    self._labelHeight,
                    lbl.symbolLabel.labelAlignments[0],
                    lbl.symbolLabel.labelOrients[0],
                    lbl.symbolLabel.labelUses[1],
                    )
                label.labelDefs()
                label.labelVisible = True
                label.setOpacity(1)
                self.symbolScene.addItem(label)
            clbFile.write("\n")


#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting or consulting/
#    support services related to the Software), a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#


import shutil

from PySide6.QtCore import (
    Qt,
)
from PySide6.QtGui import (
    QAction,
    QStandardItemModel,
)
from PySide6.QtWidgets import (
    QAbstractItemView,
    QDialog,
    QMenu,
    QMessageBox,
    QTreeView,
)


import revedaEditor.backend.libraryMethods as libm
import revedaEditor.backend.schBackEnd as scb
import revedaEditor.gui.fileDialogues as fd


class designLibrariesView(QTreeView):
    def __init__(self, parent):
        super().__init__(parent=parent)  # QTreeView
        self.parent = parent
        self.setSelectionMode(QAbstractItemView.SingleSelection)
        self.viewCounter = 0
        self.libBrowsW = self.parent.parent
        self.appMainW = self.libBrowsW.appMainW
        self.libraryDict = self.appMainW.libraryDict  # type: dict
        self.cellViews = self.appMainW.cellViews  # type: list
        self.openViews = self.appMainW.openViews  # type: dict
        self.logger = self.appMainW.logger
        self.selectedItem = None
        # library model is based on qstandarditemmodel
        self.libraryModel = designLibrariesModel(self.libraryDict)
        self.setSortingEnabled(True)
        self.setUniformRowHeights(True)
        self.expandAll()
        self.setModel(self.libraryModel)

    def removeLibrary(self):
        button = QMessageBox.question(
            self,
            "Library Deletion",
            "Are you sure to delete " "this library? This action cannot be undone.",
        )
        if button == QMessageBox.Yes:
            shutil.rmtree(self.selectedItem.data(Qt.UserRole + 2))
            self.libraryModel.removeRow(self.selectedItem.row())

    def renameLib(self):
        oldLibraryName = self.selectedItem.libraryName
        dlg = fd.renameLibDialog(self, oldLibraryName)
        if dlg.exec() == QDialog.Accepted:
            newLibraryName = dlg.newLibraryName.text().strip()
            libraryItem = libm.getLibItem(self.libraryModel, oldLibraryName)
            libraryItem.setText(newLibraryName)
            oldLibraryPath = libraryItem.data(Qt.UserRole + 2)
            newLibraryPath = oldLibraryPath.parent.joinpath(newLibraryName)
            oldLibraryPath.rename(newLibraryPath)

    def createCell(self):
        dlg = fd.createCellDialog(self, self.libraryModel)
        assert isinstance(self.selectedItem, scb.libraryItem)
        dlg.libNamesCB.setCurrentText(self.selectedItem.libraryName)
        if dlg.exec() == QDialog.Accepted:
            cellName = dlg.cellCB.currentText()
            if cellName.strip() != "":
                scb.createCell(self, self.libraryModel, self.selectedItem, cellName)
            else:
                self.logger.error("Please enter a cell name.")

    def copyCell(self):
        dlg = fd.copyCellDialog(self, self.libraryModel, self.selectedItem)

        if dlg.exec() == QDialog.Accepted:
            scb.copyCell(
                self, dlg.model, dlg.cellItem, dlg.copyName.text(), dlg.selectedLibPath
            )

    def renameCell(self):
        dlg = fd.renameCellDialog(self, self.selectedItem)
        if dlg.exec() == QDialog.Accepted:
            scb.renameCell(self, dlg.cellItem, dlg.nameEdit.text())

    def deleteCell(self):
        try:
            shutil.rmtree(self.selectedItem.data(Qt.UserRole + 2))
            self.selectedItem.parent().removeRow(self.selectedItem.row())
        except OSError as e:
            self.logger.warning(f"Error:{e}")

    def createCellView(self):
        dlg = fd.createCellViewDialog(self, self.libraryModel, self.selectedItem)
        if dlg.exec() == QDialog.Accepted:
            viewItem = scb.createCellView(
                self.appMainW, dlg.nameEdit.text(), self.selectedItem
            )
            self.libBrowsW.createNewCellView(
                self.selectedItem.parent(), self.selectedItem, viewItem
            )

    def openView(self):
        viewItem = self.selectedItem
        cellItem = viewItem.parent()
        libItem = cellItem.parent()
        self.libBrowsW.openCellView(viewItem, cellItem, libItem)

    def copyView(self):
        dlg = fd.copyViewDialog(self, self.libraryModel)
        dlg.libNamesCB.setCurrentText(self.selectedItem.parent().parent().libraryName)
        dlg.cellCB.setCurrentText(self.selectedItem.parent().cellName)
        if dlg.exec() == QDialog.Accepted:
            if self.selectedItem.data(Qt.UserRole + 1) == "view":
                viewPath = self.selectedItem.data(Qt.UserRole + 2)
                selectedLibItem = libm.getLibItem(
                    self.libraryModel, dlg.libNamesCB.currentText()
                )
                cellName = dlg.cellCB.currentText()
                libCellNames = [
                    selectedLibItem.child(row).cellName
                    for row in range(selectedLibItem.rowCount())
                ]
                if (
                    cellName in libCellNames
                ):  # check if there is the cell in the library
                    cellItem = libm.getCellItem(
                        selectedLibItem, dlg.cellCB.currentText()
                    )
                else:
                    cellItem = scb.createCell(
                        self.libBrowsW,
                        self.libraryModel,
                        selectedLibItem,
                        dlg.cellCB.currentText(),
                    )
                cellViewNames = [
                    cellItem.child(row).viewName for row in range(cellItem.rowCount())
                ]
                newViewName = dlg.viewName.text()
                if newViewName in cellViewNames:
                    self.logger.warning(
                        "View already exists. Delete cellview and try again."
                    )
                else:
                    newViewPath = cellItem.data(Qt.UserRole + 2).joinpath(
                        f"{newViewName}.json"
                    )
                    shutil.copy(viewPath, newViewPath)
                    cellItem.appendRow(scb.viewItem(newViewPath))

    def renameView(self):
        oldViewName = self.selectedItem.viewName
        dlg = fd.renameViewDialog(self.libBrowsW, oldViewName)
        if dlg.exec() == QDialog.Accepted:
            newName = dlg.newViewNameEdit.text()
            try:
                viewPathObj = self.selectedItem.data(Qt.UserRole + 2)
                newPathObj = self.selectedItem.data(Qt.UserRole + 2).rename(
                    viewPathObj.parent.joinpath(f"{newName}.json")
                )
                self.selectedItem.parent().appendRow(scb.viewItem(newPathObj))
                self.selectedItem.parent().removeRow(self.selectedItem.row())
            except FileExistsError:
                self.logger.error("Cellview exists.")

    def deleteView(self):
        try:
            self.selectedItem.data(Qt.UserRole + 2).unlink()
            itemRow = self.selectedItem.row()
            parent = self.selectedItem.parent()
            parent.removeRow(itemRow)
        except OSError as e:
            self.logger.warning(f"Error:{e.strerror}")

    def reworkDesignLibrariesView(self, libraryDict: dict):
        """
        Recreate library model from libraryDict.
        """
        self.libraryModel = designLibrariesModel(libraryDict)
        self.setModel(self.libraryModel)
        self.libBrowsW.libraryModel = self.libraryModel

    # context menu
    def contextMenuEvent(self, event):
        menu = QMenu(self)
        try:
            index = self.selectedIndexes()[0]
        except IndexError:
            pass
        try:
            self.selectedItem = self.libraryModel.itemFromIndex(index)
            if self.selectedItem.data(Qt.UserRole + 1) == "library":
                menu.addAction("Rename Library", self.renameLib)
                menu.addAction("Remove Library", self.removeLibrary)
                menu.addAction("Create Cell", self.createCell)
            elif self.selectedItem.data(Qt.UserRole + 1) == "cell":
                menu.addAction(
                    QAction("Create CellView...", self, triggered=self.createCellView)
                )
                menu.addAction(QAction("Copy Cell...", self, triggered=self.copyCell))
                menu.addAction(
                    QAction("Rename Cell...", self, triggered=self.renameCell)
                )
                menu.addAction(
                    QAction("Delete Cell...", self, triggered=self.deleteCell)
                )
            elif self.selectedItem.data(Qt.UserRole + 1) == "view":
                menu.addAction(QAction("Open View", self, triggered=self.openView))
                menu.addAction(QAction("Copy View...", self, triggered=self.copyView))
                menu.addAction(
                    QAction("Rename View...", self, triggered=self.renameView)
                )
                menu.addAction(
                    QAction("Delete View...", self, triggered=self.deleteView)
                )
            menu.exec(event.globalPos())
        except UnboundLocalError:
            pass


class designLibrariesModel(QStandardItemModel):
    def __init__(self, libraryDict):
        self.libraryDict = libraryDict
        super().__init__()
        self.rootItem = self.invisibleRootItem()
        self.setHorizontalHeaderLabels(["Libraries"])
        self.initModel()

    def initModel(self):
        for designPath in self.libraryDict.values():
            self.populateLibrary(designPath)

    def populateLibrary(self, designPath):  # designPath: Path
        """
        Populate library view.
        """
        if designPath.joinpath("reveda.lib").exists():
            libraryItem = self.addLibraryToModel(designPath)
            cellList = [cell.name for cell in designPath.iterdir() if cell.is_dir()]
            for cell in cellList:  # type: str
                cellItem = self.addCellToModel(designPath.joinpath(cell), libraryItem)
                viewList = [
                    view.name
                    for view in designPath.joinpath(cell).iterdir()
                    if view.suffix == ".json"
                ]
                for view in viewList:
                    self.addViewToModel(designPath.joinpath(cell, view), cellItem)

    def addLibraryToModel(self, designPath):
        libraryEntry = scb.libraryItem(designPath)
        self.rootItem.appendRow(libraryEntry)
        return libraryEntry

    def addCellToModel(self, cellPath, parentItem):
        cellEntry = scb.cellItem(cellPath)
        parentItem.appendRow(cellEntry)
        return cellEntry

    def addViewToModel(self, viewPath, parentItem):
        viewEntry = scb.viewItem(viewPath)
        parentItem.appendRow(viewEntry)

class symbolViewsModel(designLibrariesModel):
    """
    Initializes the object with the given `libraryDict` and `symbolViews`.

    Parameters:
        libraryDict (dict): A dictionary containing the library information.
        symbolViews (list): A list of symbol views.

    Returns:
        None
    """

    def __init__(self, libraryDict: dict, symbolViews: list):
        self.symbolViews = symbolViews
        super().__init__(libraryDict)

    def populateLibrary(self, designPath):  # designPath: Path
        """
        Populate library view.
        """
        if designPath.joinpath("reveda.lib").exists():
            libraryItem = self.addLibraryToModel(designPath)
            cellList = [cell.name for cell in designPath.iterdir() if cell.is_dir()]
            for cell in cellList:  # type: str
                cellItem = self.addCellToModel(designPath.joinpath(cell), libraryItem)
                viewList = [
                    view.name
                    for view in designPath.joinpath(cell).iterdir()
                    if view.suffix == ".json"
                    and any(x in view.name for x in self.symbolViews)
                ]
                for view in viewList:
                    self.addViewToModel(designPath.joinpath(cell, view), cellItem)


class layoutViewsModel(designLibrariesModel):
    def __init__(self, libraryDict: dict, layoutViews: list):
        self.layoutViews = layoutViews
        super().__init__(libraryDict)

    def populateLibrary(self, designPath):  # designPath: Path
        """
        Populate library view.
        """
        if designPath.joinpath("reveda.lib").exists():
            libraryItem = self.addLibraryToModel(designPath)
            cellList = [cell.name for cell in designPath.iterdir() if cell.is_dir()]
            for cell in cellList:  # type: str
                cellItem = self.addCellToModel(designPath.joinpath(cell), libraryItem)
                viewList = [
                    view.name
                    for view in designPath.joinpath(cell).iterdir()
                    if view.suffix == ".json"
                    and any(x in view.name for x in self.layoutViews)
                ]
                for view in viewList:
                    self.addViewToModel(designPath.joinpath(cell, view), cellItem)
